#include <msp/gl/renderer.h>
#include "stereocombiner.h"
#include "stereoview.h"

using namespace std;

namespace Msp {
namespace VR {

StereoView::StereoView(unsigned w, unsigned h, const GL::Camera &c, const GL::Renderable &r, const StereoCombiner &m):
	width(w),
	height(h),
	base_camera(c),
	renderable(r),
	combiner(0)
{
	set_combiner(m);
	set_eye_spacing(0.064);
	set_strabismus(Geometry::Angle<float>::zero());
}

void StereoView::set_combiner(const StereoCombiner &c)
{
	combiner = &c;

	unsigned w = width*combiner->get_width_factor();
	unsigned h = height*combiner->get_height_factor();
	left.create_target(w, h);
	right.create_target(w, h);
}

void StereoView::set_eye_spacing(float s)
{
	eye_spacing = s;
}

void StereoView::set_strabismus(const Geometry::Angle<float> &s)
{
	strabismus = s;
}

void StereoView::setup_frame() const
{
	offset_axis = normalize(cross(base_camera.get_look_direction(), base_camera.get_up_direction()))*0.5f;

	EyeParams params;
	params.fov = combiner->get_field_of_view();
	if(params.fov==Geometry::Angle<float>::zero())
		params.fov = base_camera.get_field_of_view();

	params.aspect = base_camera.get_aspect()*combiner->get_aspect_factor();
	params.near_clip = base_camera.get_near_clip();
	params.far_clip = base_camera.get_far_clip();

	float frustum_skew = combiner->get_frustum_skew();
	float halfw = tan(params.fov/2.0f)*params.aspect;
	frustum_skew = tan(Geometry::atan<float>(frustum_skew*halfw)+strabismus)/halfw;

	left.setup_frame(base_camera, offset_axis*-eye_spacing, frustum_skew, params);
	right.setup_frame(base_camera, offset_axis*eye_spacing, -frustum_skew, params);
}

void StereoView::render() const
{
	setup_frame();
	renderable.setup_frame();
	left.render(renderable);
	right.render(renderable);
	renderable.finish_frame();
	combiner->render(left.target->color, right.target->color);
}


StereoView::RenderTarget::RenderTarget(unsigned width, unsigned height)
{
	color.set_min_filter(GL::LINEAR);
	color.set_wrap(GL::CLAMP_TO_EDGE);
	color.storage(GL::RGB, width, height);
	fbo.attach(GL::COLOR_ATTACHMENT0, color);

	depth.storage(GL::DEPTH_COMPONENT, width, height);
	fbo.attach(GL::DEPTH_ATTACHMENT, depth);
}


StereoView::Eye::Eye():
	target(0)
{ }

void StereoView::Eye::create_target(unsigned w, unsigned h)
{
	delete target;
	target = new RenderTarget(w, h);
}

void StereoView::Eye::setup_frame(const GL::Camera &base_camera, const GL::Vector3 &offset, float frustum_skew, const EyeParams &params) const
{
	camera.set_position(base_camera.get_position()+offset);
	camera.set_up_direction(base_camera.get_up_direction());
	camera.set_look_direction(base_camera.get_look_direction());

	camera.set_field_of_view(params.fov);
	camera.set_aspect(params.aspect);
	camera.set_depth_clip(params.near_clip, params.far_clip);
	camera.set_frustum_axis(frustum_skew, 0);
}

void StereoView::Eye::render(const GL::Renderable &renderable) const
{
	GL::Bind bind_fbo(target->fbo);
	target->fbo.clear(GL::COLOR_BUFFER_BIT|GL::DEPTH_BUFFER_BIT);
	GL::Renderer renderer(&camera);
	renderable.render(renderer);
}

} // namespace VR
} // namespace Msp
